"""Computes the initial active set and lambda_max

This runs before solving for the first solution in the path.

"""

from heapq import *
import os
import numpy as np
import numba
from numba import njit, prange
from .Structs import *
os.environ["NUMBA_WARNINGS"] = "1"

@njit(numba.types.Tuple((numba.float64[::1],
                         numba.float64))(numba.float64[:,
                                                       ::1],
                                         numba.float64[::1],
                                         numba.int64,
                                         numba.int64,
                                         numba.int64[::1],
                                         numba.int64[::1]),
      parallel=True,
      fastmath=True,
      cache=True)
def ComputeDots(X, r, pMain, P, binsin, balancedInd):
    bins = binsin[:]
    rXijs = np.zeros(P)
    l2Norms = np.zeros(P)
    XT = X.T.copy()
    for l in prange(pMain - 1):
        i = balancedInd[l]
        # (X[:,(l+1):].T * X[:,l]).T
        XlXjs = np.transpose(XT[(i + 1):, :] * X[:, i])
        # XlXjs[:,:] is needed because of numba  #np.linalg.norm(XlXjs, axis=0)
        l2Norms[bins[i]:bins[i + 1]] = np.sum(XlXjs[:, :] ** 2, axis=0)
        rXijs[bins[i]:bins[i + 1]] = np.dot(r.T, XlXjs)

    maxNorm = np.sqrt(np.max(l2Norms))
    rXijs = np.abs(rXijs)
    return (rXijs, maxNorm)


'''
def ComputeDots(X, r, pMain, P, binsin):
    bins = binsin[:]
    rXijs = np.zeros(P)
    l2Norms = np.zeros(P)
    XT = X.T.copy()
    for l in prange(pMain-1):
        #start = #int(l*(pMain - (l+1)/2))
        #end = #start+pMain-l-1
        XlXjs = np.transpose(XT[(l+1):,:] * X[:,l]) #(X[:,(l+1):].T * X[:,l]).T
        l2Norms[bins[l]:bins[l+1]] = np.sum(XlXjs[:,:] ** 2, axis=0) #XlXjs[:,:] is needed because of numba  #np.linalg.norm(XlXjs, axis=0)
        rXijs[bins[l]:bins[l+1]] = np.dot(r.T, XlXjs)

    maxNorm = np.sqrt(np.max(l2Norms))
    rXijs = np.abs(rXijs)
    return (rXijs, maxNorm)



'''


def InitialRun(data, kmain, kint, Params):
    # Later: Make this return currentB, currentT
    X = data.X
    y = data.y
    pMain = data.pMain
    r = y - data.ybar
    Bindices = []
    Tindices = []
    rtXi = {}
    minheapmain = []
    # In what follows, we assume beta_i = 0 => theta_{i:} = 0.
    # Can handle this later -- no need for it now.
    for i in range(pMain):
        rXi = np.abs(np.dot(r, X[:, i]))  # needed here for now..
        rtXi[i] = rXi
        if len(minheapmain) < kmain:
            heappush(minheapmain, (rXi, i))
        else:
            if minheapmain[0][0] < rXi:
                heappop(minheapmain)
                heappush(minheapmain, (rXi, i))

    minheap = []
    '''
    for i in range(pMain,p):
        l,j = D[i] # (l,j) are the main effects corresponding to i
        rXi = np.abs(np.dot(r, X[:,l]*X[:,j]))
        if len(minheap) < k:
            heappush(minheap,(rXi,D[i]))
        else:
            if minheap[0][0] < rXi:
                heappop(minheap)
                heappush(minheap,(rXi,D[i]))
    '''
    '''
    keys = {(i,j) for i in range(pMain-1) for j in range(i+1,pMain)}
    L2Norms = dict.fromkeys(keys) # allocate space for the dict initially
    for l in range(pMain-1):
        #rlTimesXl = r*X[:,l]
        #rXis = np.abs(np.dot(rlTimesXl.T, X[:,(l+1):]))
        XlXjs = (X[:,(l+1):].T * X[:,l]).T
        rXis = np.abs(np.dot(r.T, XlXjs))
        L2NormsTemp = np.linalg.norm(XlXjs, axis=0) # finds the l2 norm of every column
        for j in range(l+1, pMain):
            #rXi = rXis[j-l-1] #np.abs(np.dot(X[:,j], rlTimesXl))
            index = j-l-1
            XlXj = XlXjs[:,index]
            rXi = rXis[index] #np.abs(np.dot(r.T,XlXj))
            if len(minheap) < kint:
                heappush(minheap,(rXi,(l,j)))
            else:
                if minheap[0][0] < rXi:
                    heappop(minheap)
                    heappush(minheap,(rXi,(l,j)))
            L2Norms[(l,j)] = L2NormsTemp[index] #np.sqrt(np.dot(XlXj,XlXj))
    maxNorm = max(L2Norms.values())
    data.maxNorm = maxNorm
    '''
    rXijs, maxNorm = ComputeDots(
        X, r, pMain, data.P, data.bins, data.balancedInd)
    data.maxNorm = maxNorm
    topIndicesFlat = np.argpartition(rXijs, -kint)[-kint:]
    topValues = rXijs[topIndicesFlat]
    topIndices = FlatToTuple(topIndicesFlat, data)
    minheap = list(zip(topValues, topIndices))

    for pair in sorted(minheapmain, reverse=True):
        i = pair[1]
        if i not in Bindices:
            Bindices.append(i)

    for pair in sorted(minheap, reverse=True):
        i = pair[1][0]
        j = pair[1][1]
        if i not in Bindices:
            Bindices.append(i)
        if j not in Bindices:
            Bindices.append(j)
        #index = int((i+1)*(pMain-1) - i*(i-1)/2 + (j-i))
        # if index not in Tindices:
        #    Tindices.append(index)
        if (i, j) not in Tindices:
            Tindices.append((i, j))

    #Bindices = sorted(Bindices)
    #Tindices = sorted(Tindices)

    # lambdaMax = max(max(minheapmain)[0], max(minheap)[0])*Params["M"] #
    # *Params["M"] since lambdaMax is lambda0Max

    # *Params["M"] since lambdaMax is lambda0Max
    lambdaMax = max(minheapmain)[0] * Params["M"]

    return (Bindices, Tindices, lambdaMax, maxNorm)



""" Checks the KKT conditions outside of the active set. """

from heapq import *
import os
import time
import warnings
import numpy as np
import numba
from numba import njit, prange
import networkx as nx
from gurobipy import *
from .Structs import *

from numba.core.errors import NumbaPerformanceWarning
os.environ["NUMBA_WARNINGS"] = "1"
warnings.filterwarnings("ignore", category=NumbaPerformanceWarning)


@njit(numba.float64[::1](numba.float64[:,
                                       ::1],
                         numba.float64[::1],
                         numba.int64,
                         numba.int64,
                         numba.int64[::1],
                         numba.int64[::1]),
      parallel=True,
      fastmath=True,
      cache=True)
def ComputeDots(X, r, pMain, P, binsin, balancedInd):
    bins = binsin[:]  # used for better cache locality.
    rlTimesX = X.T * r
    rXijs = np.zeros(P)
    for l in prange(pMain - 1):
        i = balancedInd[l]
        # bins are used for load balancing.
        rXijs[bins[i]:bins[i + 1]] = np.dot(rlTimesX[i, :], X[:, (i + 1):])
    return rXijs

def CheckViolating(
        X,
        r,
        fixedSet,
        fixedZeroSet,
        Binnz,
        minheap,
        Tindices,
        Tviolating,
        Lambda1ovM):
    '''
    fixedSet.remove((-1,-1)) # hack for numba's empty list/set problem -- remove when numba fixes this
    fixedZeroSet.remove(-1)
    Binnz.remove(-1)
    minheap.pop()
    Tindices.pop()
    Tviolating.remove((-1,-1))
    '''
    feasible = True
    for l, j in Tviolating:
        rXi = np.abs(np.dot(r, X[:, l] * X[:, j]))
        Blnnz = l in Binnz
        # is either of l or j fixed to zero?
        fixedlj0 = (l in fixedZeroSet) or (j in fixedZeroSet)
        if ((l, j) not in fixedSet) and not fixedlj0:
            # can be improved later by splitting into several cases.
            if (not Blnnz) or (j not in Binnz):
                if rXi > Lambda1ovM:
                    minheap.append((rXi, (l, j)))
            # l and j are both nnz (and not fixed to zero).
            elif rXi > Lambda1ovM and (l, j) not in Tindices:
                # OK since l and j are already in activeset.
                Tindices.append((l, j))
                feasible = False
    return feasible


'''
def ComputeDots(X, r, pMain, P, binsin):
    bins = binsin[:] # used for better cache locality
    rlTimesX = X.T * r
    rXijs = np.zeros(P)
    for l in prange(pMain-1):
        rXijs[bins[l]:bins[l+1]] = np.dot(rlTimesX[l,:], X[:,(l+1):])
    return rXijs
'''

'''
@njit(parallel=True, fastmath=True, cache=True)
def ComputeDots(X, r, pMain, P):
    rlTimesX = X.T * r
    rXijs = np.zeros(P)
    for l in prange(pMain-1):
        start = int(l*(pMain - (l+1)/2))
        rXijs[start:(start+pMain-l-1)] = np.dot(rlTimesX[l,:], X[:,(l+1):])
    return rXijs
'''


def CheckOptScreen(
        data,
        Bindices,
        Tindices,
        currentB,
        currentT,
        r,
        rold,
        maxrtX,
        Tviolating,
        maxNorm,
        Params,
        BscreenIndices,
        fixedVars):
    '''
    Finds Bi's and Tij's (outside the Bindices and Tindices) that violate the optimality
    conditions. Main logic:
        What not to check?
        - If i in fixedVars, do not check Bi
        - If (i in fixedVars and zero) or (j in fixedVars and zero) then do not check Tij (since any nnz Tij is infeasible)

        What to check?
        - If i not in fixedVars and not in Bindices, check Bi
        - If Bi or Bj is zero, check Tij
        - If Bi and Bj are nnz, but (i,j) not in Tindices, check Tij

    Only a portion of the violations in each subgraph is appended to Bindices/Tindices (current threshold = 5) for each subgraph
    The appended Bindices/Tindices are sorted by decreasing order of |<r,Xi>|.

    Bindices and Tindices are sets of indices that are (potentially) modified.

    ToDo: Fix BscreenIndices (not an issue for now since we're using BscreenIndices = range(pMain))
    '''
    X = data.X
    pMain = data.pMain
    #D = data.D
    feasible = True  # stays true unless an infeasbility is found
    slackimp = True
    # set k to a relatively large number.
    k = 1000000
    M = Params["M"]
    Lambda0 = Params["Lambda"]
    Lambda1 = Params["alpha"] * Lambda0
    MovLambda = M / Lambda0
    LambdaovM = Lambda0 / M
    Lambda1ovM = Lambda1 / M

    Bind = {}
    for i in range(len(Bindices)):
        Bind[Bindices[i]] = i

    #p = pMain + len(D)
    rtXi = {}
    minheapmain = []

    # H. asm. below.
    # BCheckSet contains the indices of all the Bi's not in fixedVars (i.e.,
    # typically, [p] - fixedVars).
    BCheckSet = set(BscreenIndices).difference(set(fixedVars.keys()))
    for i in BCheckSet:
        Bizero = i not in Bindices or (
            i in Bindices and currentB[Bind[i]] == 0)
        if Bizero:
            rXi = np.abs(np.dot(r, X[:, i]))  # needed here for now.
            rtXi[i] = rXi
            if len(minheapmain) <= k:
                heappush(minheapmain, (rXi, i))
            else:
                if minheapmain[0][0] < rXi:
                    heappop(minheapmain)
                    heappush(minheapmain, (rXi, i))

    multiplierTimesLambda1ovM = 0.7 * Lambda1ovM
    firstRun = maxrtX is None
    if not firstRun:
        gamma = np.dot(rold, r) / np.dot(rold, rold)  # 1
        tempdiff = r - gamma * rold
        eps = np.sqrt(np.dot(tempdiff, tempdiff))
        #norm = max(L2Norms.values())
        Params["print"]("maxrtX: ", maxrtX)
        Params["print"]("(LambdaovM - eps*norm)/gamma: ",
                        (LambdaovM - eps * data.maxNorm) / np.abs(gamma))

    minheap = []

    '''
    Binnz = {} # indicates whether Bi is (nnz or in fixedVars) = True, or not (False)
    for i in BscreenIndices:
        Binnz[i] = (i in Bindices and currentB[Bind[i]] != 0) or (i in fixedVars)
    if not firstRun and maxrtX <= (LambdaovM - eps*data.maxNorm)/np.abs(gamma):
        Params["print"]("################## !!!!!!!!!!!!!!!! #####################")
        Params["print"]("################## !!!!!!!!!!!!!!!! #####################")
        Params["print"]("################## !!!!!!!!!!!!!!!! #####################")
        Params["print"]("################## !!!!!!!!!!!!!!!! #####################")
        Params["print"]("Avoided full evaluation!")

        for l,j in Tviolating:
            rXi = np.abs(np.dot(r, X[:,l]*X[:,j]))
            Blnnz = Binnz[l]
            # is either of l or j fixed to zero?
            fixedlj0 = (l in fixedVars and fixedVars[l] == 0) or (j in fixedVars and fixedVars[j] == 0)
            if ((l,j) not in fixedVars) and not fixedlj0:
                if (not Blnnz) or (not Binnz[j]): # can be improved later by splitting into several cases
                    if rXi > LambdaovM:
                        minheap.append((rXi,(l,j)))
                else: # l and j are both nnz (and not fixed to zero)
                    if rXi > LambdaovM and (l,j) not in Tindices: # add appx kkt here later
                        Tindices.append((l,j)) # OK since l and j are already in activeset.
                        Params["print"]("!!!!!!")
                        Params["print"]("!!!!!!")
                        Params["print"]("!!!!!! Appended T", (l,j))
                        Params["print"]("rXi: ", rXi, " LambdaovM: ", LambdaovM)
                        feasible = False
    '''

    Binnz = {i for i in BscreenIndices if (
        i in Bindices and currentB[Bind[i]] != 0) or (i in fixedVars)}

    if not firstRun and maxrtX <= (
            Lambda1ovM - eps * data.maxNorm) / np.abs(gamma):
        Params["print"](
            "################### !!!!!!!!!!!!!!!! #####################")
        Params["print"]("Avoided full evaluation!")

        start = time.time()
        fixedZeroSet = {
            key for key,
            val in fixedVars.items() if not isinstance(
                key,
                tuple) and val == 0}
        fixedSet = {key for key in fixedVars if isinstance(key, tuple)}
        end = time.time()
        Params["print"]("!!! Time for precheck: ", end - start)
        '''
        minheap.append((-1.1,(-1,-1)))
        Tindices.append((-1,-1))
        fixedSet.add((-1,-1))
        fixedZeroSet.add(-1)
        Binnz.add(-1)
        Tviolating.add((-1,-1))
        '''
        start = time.time()
        feasible = CheckViolating(
            X,
            r,
            fixedSet,
            fixedZeroSet,
            Binnz,
            minheap,
            Tindices,
            Tviolating,
            Lambda1ovM)
        end = time.time()
        Params["print"]("!!! Time for avoidance check: ", end - start)

    else:
        # '''
        rold = r

        rXijs = ComputeDots(
            X,
            r,
            data.pMain,
            data.P,
            data.bins,
            data.balancedInd)
        rXijs = np.abs(rXijs)

        ViolatingIndices = np.nonzero(rXijs > multiplierTimesLambda1ovM)[0]
        OptCheckIndices = np.nonzero(rXijs > Lambda1ovM)[0]

        #sortedIndices = np.argsort(rXijs)
        #np.searchsorted(rXijs, [], side="right", sorter=sortedIndices)

        # Remove the indices of (i,j)'s that are fixed or one of i and j is
        # fixed to zero.
        skipIndices = []
        for key in fixedVars:
            if not isinstance(key, tuple) and fixedVars[key] == 0:
                skipIndices = skipIndices + \
                    [(key, j) for j in range(key + 1, pMain)] + [(i, key) for i in range(0, key)]
            elif isinstance(key, tuple) and fixedVars[key] == 0:
                skipIndices.append(key)
        for counter, (i, j) in enumerate(skipIndices):
            skipIndices[counter] = int(
                (i + 1) * (pMain - 1) - i * (i - 1) / 2 + (j - i) - pMain)

        if len(skipIndices) != 0:
            ViolatingIndices = np.setdiff1d(ViolatingIndices, skipIndices)
            OptCheckIndices = np.setdiff1d(OptCheckIndices, skipIndices)

        Tviolating = set(FlatToTuple(ViolatingIndices, data))
        minheapindices = FlatToTuple(OptCheckIndices, data)

        minheapTemp = list(zip(rXijs[OptCheckIndices], minheapindices))
        for val, (i, j) in minheapTemp:
            if i in Binnz and j in Binnz:
                if (i, j) not in Tindices:
                    # OK since l and j are already in activeset.
                    Tindices.append((i, j))
                    Params["print"]("!!!!!!")
                    Params["print"]("!!!!!!")
                    Params["print"]("!!!!!! Appended T", (i, j))
                    Params["print"]("rXi: ", val, " Lambda1ovM: ", Lambda1ovM)
                    feasible = False
            else:
                minheap.append((val, (i, j)))

        mask = np.full(data.P, False)
        mask[ViolatingIndices] = True
        mask[skipIndices] = True
        rXijsMasked = np.ma.array(rXijs, mask=mask)  # copies by reference
        maxrtX = np.ma.max(rXijsMasked)
        # '''

    # Find the connected components

    start = time.time()
    G = nx.Graph()
    G.add_edges_from(v[1] for v in minheap)
    disconnectedGraphs = (G.subgraph(c).edges()
                          for c in nx.connected_components(G))
    end = time.time()
    Params["print"]("!!! Time for constructing graphs: ", end - start)

    # Check feasibility of the connected LPs.
    rtXij = {v[1]: v[0] for v in minheap}
    infeasibleGraphs = []
    for graph in disconnectedGraphs:
        Params["print"]("Graph Size: ", len(graph))
        lpModel = Model()
        lpModel.setParam('OutputFlag', False)
        mu = {}
        psi = {}
        for i, j in graph:
            izero = i not in Bindices or (
                i in Bindices and currentB[Bind[i]] == 0)
            izero = izero and (i not in fixedVars)

            jzero = j not in Bindices or (
                j in Bindices and currentB[Bind[j]] == 0)
            jzero = jzero and (j not in fixedVars)

            if izero:
                if i not in mu:
                    mu[i] = {}
                mu[i][j] = lpModel.addVar(vtype=GRB.CONTINUOUS, lb=0)

            if jzero:
                if j not in mu:
                    mu[j] = {}
                mu[j][i] = lpModel.addVar(vtype=GRB.CONTINUOUS, lb=0)

            if slackimp:
                psi[(i, j)] = lpModel.addVar(vtype=GRB.CONTINUOUS, lb=0)

            if izero and not jzero:
                lpModel.addConstr(mu[i][j] == (
                    M / Lambda0) * rtXij[(min(i, j), max(i, j))] - Lambda1 / Lambda0 - psi[(i, j)])

            elif jzero and not izero:
                lpModel.addConstr(mu[j][i] == (
                    M / Lambda0) * rtXij[(min(i, j), max(i, j))] - Lambda1 / Lambda0 - psi[(i, j)])

            elif izero and jzero:
                lpModel.addConstr(mu[i][j] +
                                  mu[j][i] == (M /
                                               Lambda0) *
                                  rtXij[(min(i, j), max(i, j))] -
                                  Lambda1 /
                                  Lambda0 -
                                  psi[(i, j)])

        for i in mu:
            if not slackimp:
                lpModel.addConstr(
                    quicksum(mu[i].values()) <= 1 - (M / Lambda0) * rtXi[i])
            else:
                if (M / Lambda0) * rtXi[i] <= 1:
                    lpModel.addConstr(
                        quicksum(mu[i].values()) - 1 + (M / Lambda0) * rtXi[i] <= 0)

        if not slackimp:
            lpModel.setObjective(0, GRB.MINIMIZE)
            lpModel.update()
            lpModel.optimize()
            if lpModel.Status == 3:
                infeasibleGraphs.append(graph)
                lpModel.feasRelaxS(0, False, False, True)
                Params["print"]("Optimizing for infeasibility")
                lpModel.optimize()
                obj = lpModel.getObjective()
                if obj.getValue() > 0.05:
                    feasible = False
                else:
                    Params["print"]("Opt. conditions appx. satisfied")
                # break
        else:
            eta = lpModel.addVar(vtype=GRB.CONTINUOUS, lb=0)
            for (i, j) in psi:
                lpModel.addConstr(eta >= psi[(i, j)])
            lpModel.setObjective(eta, GRB.MINIMIZE)
            lpModel.update()
            lpModel.optimize()
            obj = lpModel.getObjective()
            if obj.getValue() > 0.05:
                feasible = False
                infeasibleGraphs.append(graph)
            else:
                Params["print"]("Opt. conditions satisfied")

    Params["print"]("Feasible Graphs?: ", feasible)

    # Keep the following after checking graph feasibility (since graph feas.
    # depends on current Bindices).
    threshold = 10
    minheap = []
    for i in rtXi:
        if len(minheap) < threshold:
            heappush(minheap, (rtXi[i], i))
        else:
            if minheap[0][0] < rtXi[i]:
                heappop(minheap)
                heappush(minheap, (rtXi[i], i))
    for v in sorted(minheap, reverse=True):
        i = v[1]
        if (M / Lambda0) * rtXi[i] >= 1:
            if i not in Bindices:
                feasible = False
                Bindices.append(i)
                Params["print"]("!!!!!!")
                Params["print"]("!!!!!!")
                Params["print"]("!!!!!! Appended B", i)
                Params["print"]("rXi: ", rtXi[i], " LambdaovM: ", LambdaovM)

    Params["print"]("Feasible? ", feasible)
    if not feasible:
        for graph in infeasibleGraphs:
            minheap = []
            for pair in graph:
                i = min(pair[0], pair[1])
                j = max(pair[0], pair[1])
                mag = rtXij[(i, j)]
                if (i, j) not in Tindices:
                    if len(minheap) < threshold:
                        heappush(minheap, (mag, (i, j)))
                    else:
                        if minheap[0][0] < mag:
                            heappop(minheap)
                            heappush(minheap, (mag, (i, j)))

            for v in sorted(minheap, reverse=True):
                pair = v[1]
                i = min(pair[0], pair[1])
                j = max(pair[0], pair[1])
                # index not in Tindices by construction of minheap.
                Tindices.append((i, j))
                if i not in Bindices:
                    Bindices.append(i)
                if j not in Bindices:
                    Bindices.append(j)

    return (feasible, Bindices, Tindices, rold, maxrtX, Tviolating, maxNorm)



"""Runs the solver for a single lambda in the path."""
from .RelaxationSolver import PGD
from .CheckOptScreen import CheckOptScreen
from .Structs import *


def FitSingleLambda(Params, relaxationVars, fixedVars, data):
    """
    Fits the convex relaxation of the strong hierarchy model for a single value of the regularization parameter lambda.

    Parameters:
        Params (dict): A dictionary of parameters with the following keys
            fixedVars (dict)
            DualInitial (bool): indicates whether to use dual initialization on the first call to PGD.
            Bindices (list): list of the Bi's to consider in the active set.
            Tindices (list): list of the Tij's to consider in the active set. Each element is of the form (i,j).
            currentB (np.array): coefficient values corresponding to Bindices.
            currentT (np.array): coefficient values corresponding to Tindices.
            u (dict): dictionary of the dual variables corresponding to Bi's. key = index (i.e., i) and value = coefficient value.
            w (dict): dictionary of the dual variables corresponding to Tij's. key = (i,j) and value = coefficient value.
            rold (np.array): the vector of residuals at which the last **full** check of the optimality conditions was done.
            maxrtX (float): the maximum of r'X at the last full check of the opt. conditions.
            Tviolating (list): the indices of the violating Tij's in the last full check of optimality conditions.
            maxNorm: the max l2 norm of the non-violating columns in the last full check of optimality conditions.

    Output: A tuple (B, T, objective, integral)
            Also, Params will be modified such that Bindices, Tindices, currentB, currentT, u, v, rold, maxrtX, Tviolating, and maxNorm reflect
            the optimal solution obtained at the end of the fitting process.
    """

    # Split fixedVars into fixedB and fixedT
    fixedB = {
        key: val for key,
        val in fixedVars.items() if not isinstance(
            key,
            tuple)}  # == int can cause problems
    fixedT = {
        key: val for key,
        val in fixedVars.items() if isinstance(
            key,
            tuple)}
    #Params["print"]("fixedVars: ", fixedVars)
    #Params["print"]("fixedB: ", fixedB)
    #Params["print"]("fixedT: ", fixedT)
    Bindices = relaxationVars.BActive
    Tindices = relaxationVars.TActive

    feasible = False
    while not feasible:
        sol, solTrunc, obj, integral, r, u, w = PGD(
            Params, relaxationVars, fixedB, fixedT, data)
        ###
        # This block is for debugging only
        '''
        B, T, obj, integral, r = RunRelaxation(Params, Bindices.copy(), Tindices.copy(), fixedVars, data)
        sol = Solution(B,T)
        u = {} ####
        w = {} ####
        '''
        ###
        currentB, currentT = sol.ToArray(Bindices, Tindices)

        oldBindices = Bindices.copy()
        oldTindices = Tindices.copy()

        # New imp using CheckOptScreen instead of CheckOpt.
        feasible, Bindices, Tindices, rold, maxrtX, Tviolating, maxNorm = CheckOptScreen(data, Bindices.copy(), Tindices.copy(
        ), currentB, currentT, r, relaxationVars.rold, relaxationVars.maxrtX, relaxationVars.Tviolating, relaxationVars.maxNorm, Params, range(data.pMain), fixedVars)
        relaxationVars.rold = rold
        relaxationVars.maxrtX = maxrtX
        relaxationVars.Tviolating = Tviolating
        relaxationVars.maxNorm = maxNorm

        # update currentB and currentT (since Bindices and Tindices might get
        # updated).
        currentB, currentT = sol.ToArray(Bindices, Tindices)

        if not feasible:
            Params["print"](
                "Optimality conditions not satisfied outside active set!!!")

            additionalBindices = list(
                set(Bindices).difference(
                    set(oldBindices)))
            additionalTindices = list(
                set(Tindices).difference(
                    set(oldTindices)))

            if len(additionalBindices) + len(additionalTindices) == 0:
                Params["print"]("Obtained same supp. Skipping...")
                break

            # force using dual init after the first call.
            relaxationVars.useDual = True

            Params["print"]("Augmenting Data...")
            data.Augment(additionalBindices, additionalTindices)

        relaxationVars.BActive = Bindices
        relaxationVars.TActive = Tindices
        relaxationVars.initialSol = sol
        relaxationVars.u = u
        relaxationVars.w = w

    return (solTrunc, obj, integral)



"""Implementation of Proximal Gradient Descent with Dual Block Coordinate Ascent.
Proximal screening rules are employed to identify zero interaction coordinates
and blocks.
"""

import time
import math
from copy import copy
from collections import defaultdict
import numpy as np
from gurobipy import *
import numba
from numba import njit
from .Structs import *


@njit(
    numba.float64(
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64),
    fastmath=True,
    cache=True)
def T(
        wij,
        wji,
        thetaij,
        M,
        Lambda0,
        L,
        frac,
        frac1,
        Mpfrac1,
        LambdaovM,
        Lambda1ovM):
    h = thetaij - frac * (wij + wji)
    habs = np.abs(h)
    if habs <= frac1:
        out = 0
    elif habs <= Mpfrac1:
        out = (habs - frac1) * np.sign(h)
    elif h >= Mpfrac1:
        out = M
    elif h <= -Mpfrac1:
        out = -M
    return (L / 2) * (out - thetaij)**2 + LambdaovM * \
        out * (wij + wji) + Lambda1ovM * np.abs(out)


@njit(
    numba.float64(
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64),
    fastmath=True,
    cache=True)
def q(ui, betai, M, Lambda0, L, frac):
    #frac = Lambda0/(M*L)
    if ui >= (betai - M) / frac and ui <= (betai + M) / frac:
        out = - frac * frac * L * ui * ui / 2 + (Lambda0 / M) * betai * ui
    elif ui <= (betai - M) / frac:
        out = (L / 2) * (M - betai)**2 + Lambda0 * ui
    elif ui >= (betai + M) / frac:
        out = (L / 2) * (M + betai)**2 - Lambda0 * ui
    return out


@njit(
    numba.float64(
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64),
    fastmath=True,
    cache=True)
def delT(wij, wji, thetaij, M, Lambda0, L, frac, frac1, Mpfrac1, LambdaovM):
    h = thetaij - frac * (wij + wji)
    habs = np.abs(h)
    if habs <= frac1:
        out = 0
    elif habs <= Mpfrac1:
        out = (habs - frac1) * np.sign(h) * LambdaovM
    elif h >= Mpfrac1:
        out = Lambda0
    elif h <= -Mpfrac1:
        out = -Lambda0
    return out


@njit(
    numba.float64(
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64),
    fastmath=True,
    cache=True)
def delq(ui, betai, M, Lambda0, L, frac, Mpfrac, fracsqL, LambdaovM):
    #frac = Lambda0/(M*L)
    if ui >= (betai - M) / frac and ui <= (betai + M) / frac:
        out = - fracsqL * ui + LambdaovM * betai
    elif ui <= (betai - M) / frac:
        out = Lambda0
    elif ui >= (betai + M) / frac:
        out = -Lambda0
    return out


@njit(
    numba.float64(
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64),
    fastmath=True,
    cache=True)
def dualtoprimalu(ui, betai, M, Lambda0, L, frac):
    #frac = Lambda0/(M*L)
    if ui >= (betai - M) / frac and ui <= (betai + M) / frac:
        out = betai - frac * ui
    elif ui <= (betai - M) / frac:
        out = M
    elif ui >= (betai + M) / frac:
        out = -M
    return out


@njit(
    numba.float64(
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64,
        numba.float64),
    fastmath=True,
    cache=True)
def dualtoprimalw(wij, wji, thetaij, M, Lambda0, L, frac, frac1, Mpfrac1):
    #frac = Lambda0/(M*L)
    h = thetaij - frac * (wij + wji)
    habs = np.abs(h)
    if habs <= frac1:
        out = 0
    elif habs <= Mpfrac1:
        out = (habs - frac1) * np.sign(h)
    elif h >= Mpfrac1:
        out = M
    elif h <= -Mpfrac1:
        out = -M
    return out


def terminate(oldObj, newObj, tol):
    # handle the case of all zeros optimal solution.
    if oldObj - newObj == 0:
        return True
    elif oldObj == -math.inf or oldObj == math.inf:
        return False
    elif np.abs(oldObj - newObj) / oldObj < tol:
        return True
    else:
        return False


@njit(numba.float64[::1](numba.float64[::1]), fastmath=True, cache=True)
def project(v):
    # projects on the l1-ball
    vmod = np.abs(v)
    nz = vmod[vmod.nonzero()]
    s = np.sort(nz)[::-1]
    cs = s.cumsum()
    if cs[-1] <= 1:
        return v
    ks = 1 + np.arange(len(s))
    idx = np.searchsorted(s * ks <= cs - 1, True) - 1
    lmbda = (cs[idx] - 1) / (1 + idx)
    thresh_mult = np.maximum(1 - lmbda / vmod, 0)
    return thresh_mult * v


def PGD(Params, relaxationVars, fixedBs, fixedTs, data):
    """
    Runs proximal gradient descent to solve the interactions problem on a specified active set.
    The inputs are not modified by PGD.
    """
    Tol = Params["tol"]
    TolCD = Params["tolCD"]
    Lambda0 = Params["Lambda"]
    Lambda1 = Params["alpha"] * Lambda0
    M = Params["M"]
    y = data.ycentered  # data.y - data.ybar

    Bindices = relaxationVars.BActive.copy()  # list
    Tindices = relaxationVars.TActive.copy()  # list of tuples (i,j)
    currentB, currentT = relaxationVars.initialSol.ToArray(Bindices, Tindices)
    fixedB = fixedBs.copy()  # Dict. key = index, value = 0 or 1 (no index if not fixed)
    fixedT = fixedTs.copy()  # Dict. key = (i,j), value = 0 or 1 (no index if not fixed)
    DualInitial = relaxationVars.useDual

    # Store the index mappings
    Bmap = {}  # Bmap[i] = index of i in currentB or XB
    for i in range(len(Bindices)):
        Bmap[Bindices[i]] = i

    Tmap = {}  # Tmap[(i,j)] = index of interaction in XT and currentT
    for i in range(len(Tindices)):
        c1, c2 = Tindices[i]
        Tmap[(c1, c2)] = i
        Tmap[(c2, c1)] = i

    # Next: Some sanity checks (those can be removed if we're carful about the
    # inputs)

    # Make sure if B_i is fixed to 0 then all T_{ij}'s (in Tindices) are also
    # fixed to zero
    for i, val in fixedB.items():
        if val == 0:
            for l, j in Tmap:
                if l < j and (l == i or j == i):
                    fixedT[(l, j)] = 0

    # Make sure if T_{ij} is fixed to 1 then both B_i and B_j are fixed to 1
    for key, val in fixedT.items():
        if val == 1:
            i, j = key
            fixedB[i] = 1
            fixedB[j] = 1

    # Delete from Bindices and Tindices all the indices s.t. z_i = 0 / z_{ij}
    # = 0
    Bzeros = []
    for i, val in fixedB.items():
        if val == 0:
            Bzeros.append(Bmap[i])
    for i in sorted(Bzeros, reverse=True):
        del Bindices[i]
    currentB = np.delete(currentB, Bzeros)

    Tzeros = []
    for key, val in fixedT.items():
        if val == 0:
            Tzeros.append(Tmap[key])
    for i in sorted(Tzeros, reverse=True):
        del Tindices[i]
    currentT = np.delete(currentT, Tzeros)

    # Update the index mappings
    Bmap = {}  # Bmap[i] = index of i in currentB or XB
    for i in range(len(Bindices)):
        Bmap[Bindices[i]] = i

    Tmap = {}  # Tmap[(i,j)] = index of interaction in XT and currentT
    for i in range(len(Tindices)):
        c1, c2 = Tindices[i]
        Tmap[(c1, c2)] = i
        Tmap[(c2, c1)] = i

    # End of sanity checks

    # Retrive the matrices of the optimization variables
    # Later: We can store the centered columns (but this will require twice
    # the memory)
    XB, XT = data.Retrieve(Bindices, Tindices)
    XBMean = XB.mean(axis=0)
    XB = XB - XBMean
    XTMean = XT.mean(axis=0)
    XT = XT - XTMean

    Bfree = [i for i in Bindices if i not in fixedB]
    Tfree = [(i, j) for i, j in Tmap if i < j and (i, j) not in fixedT]
    TfreeIndices = [Tmap[(i, j)]
                    for i, j in Tmap if i < j and (i, j) not in fixedT]
    lenFixedB = len(Bindices) - len(Bfree)
    lenFixedT = len([key for key in fixedT if fixedT[key] == 1])

    # (Dual) Block CD Variables
    u = defaultdict(float)
    w = defaultdict(dict)
    if not DualInitial:
        for i in Bindices:
            u[i] = 0
        for pair in Tmap:
            i, j = pair
            w[i][j] = 0
    else:
        for i in Bindices:
            if i in relaxationVars.u and i not in fixedB:
                u[i] = relaxationVars.u[i]
            else:
                u[i] = 0
        for i, j in Tmap:
            if j in relaxationVars.w[i] and (min(i, j), max(
                    i, j)) not in fixedT and i not in fixedB and j not in fixedB:
                w[i][j] = relaxationVars.w[i][j]
            else:
                # Important: we need w[i][j] = 0 if T_{ij} if fixed (this is
                # due to the thresholding function)
                w[i][j] = 0

    sortedIndices = {i: sorted(w[i]) for i in w}
    sortedIndices = defaultdict(list, sortedIndices)

    # Prepare all the fixed matrices/vectors required for grad evaluation
    # later.
    XBty = np.dot(XB.T, y)
    XBtXB = np.dot(XB.T, XB)
    XTty = np.dot(XT.T, y)
    XTtXT = np.dot(XT.T, XT)
    XBtXT = np.dot(XB.T, XT)

    # Compute the lipschitz constant of the grad.
    Xfull = np.hstack((XB, XT))
    if Xfull.shape[1] != 0:
        eigvals, v = np.linalg.eig(np.dot(Xfull.T, Xfull))
        L = np.max(np.real(eigvals))
    else:
        L = 1  # any value here should suffice - it's not used.

    # Compute the lipschitz constants for BCD.
    LCD = {}
    for i in Bindices:
        LCD[i] = (len(w[i]) + 1) * ((Lambda0**2) / (L * M**2))

    # Define the thresholding constants
    frac = Lambda0 / (M * L)
    Mpfrac = M + frac
    frac1 = Lambda1 / (M * L)
    Mpfrac1 = M + frac1
    fracsqL = frac * frac * L
    LambdaovM = Lambda0 / M
    Lambda1ovM = Lambda1 / M
    Lambda1ovLambda0 = Lambda1 / Lambda0

    start = time.time()

    oldObj = math.inf
    for it in range(5000):
        grad_B = - XBty + np.dot(XBtXB, currentB) + np.dot(XBtXT, currentT)
        grad_T = - XTty + np.dot(XTtXT, currentT) + np.dot(XBtXT.T, currentB)
        Bstar = currentB - grad_B / L
        Tstar = currentT - grad_T / L
        # Iterate over the blocks, running dual BCD.
        # We employ dual warm starts by using the same (u,w) across the PGD updates.
        CDPrevObj = -math.inf
        LCDCurrent = copy(LCD)
        useZeroSuffCondition = True
        if useZeroSuffCondition:
            # Perform proximal screening below.
            zeroGroups = set()
            for i in Bfree:
                zeroSufficient = False
                cumsum = 0
                for j in w[i]:
                    thrshld = max(
                        (abs(Tstar[Tmap[(i, j)]]) / frac - Lambda1ovLambda0), 0)
                    # Do feature level screening below.
                    if thrshld == 0:
                        # The initialization below ensures that \theta_{ij} is
                        # never updated by BCA.
                        w[i][j] = 0
                        w[j][i] = 0
                    else:
                        cumsum += thrshld

                if cumsum <= 1 - abs(Bstar[Bmap[i]]) / frac:
                    zeroSufficient = True
                if zeroSufficient:
                    u[i] = Bstar[Bmap[i]] / frac
                    for j in w[i]:
                        if abs(Tstar[Tmap[(i, j)]]) > frac1:
                            w[i][j] = Tstar[Tmap[(
                                i, j)]] / frac - Lambda1ovLambda0 * np.sign(Tstar[Tmap[(i, j)]])
                        else:
                            w[i][j] = 0
                        w[j][i] = 0
                        # Not nec. but can improve speed.
                        LCDCurrent[j] -= (Lambda0**2) / (L * M**2)
                    zeroGroups.add(i)

            BfreeMinusZeroGroups = [i for i in Bfree if i not in zeroGroups]
            CDObjConst = 0
            '''
            for i in zeroGroups:
                CDObjConst += q(u[i], Bstar[Bmap[i]], M, Lambda0, L,frac)
                for j in w[i]:
                    if i < j:
                        # T(wij, wji, thetaij, M, Lambda0, L, frac, frac1, Mpfrac1, LambdaovM, Lambda1ovM)
                        CDObjConst += T(w[i][j], w[j][i], Tstar[Tmap[(i,j)]], M, Lambda0, L,frac, frac1, Mpfrac1, LambdaovM, Lambda1ovM)
            '''
            ####
        else:
            zeroGroups = set()
            CDObjConst = 0
            BfreeMinusZeroGroups = Bfree
        # To Turn the part above off, comment it out and set the following:
        # zeroGroups = set()
        # CDObjConst = 0
        # BfreeMinusZeroGroups = Bfree

        for innerit in range(10000):
            # for i in Bfree:
            for i in BfreeMinusZeroGroups:
                # First, Calculate utilde and wtilde for ith block
                utilde = u[i] + delq(u[i],
                                     Bstar[Bmap[i]],
                                     M,
                                     Lambda0,
                                     L,
                                     frac,
                                     Mpfrac,
                                     fracsqL,
                                     LambdaovM) / LCDCurrent[i]

                #wtilde = {}
                # for j in w[i]:
                # if B_j is fixed to 1, then we already set w[j][i] = 0
                #    wtilde[j] = w[i][j] + delT(w[i][j], w[j][i], Tstar[Tmap[(i,j)]], M, Lambda0, L,frac,  Mpfrac, fracsqL, LambdaovM)/LCD[i]
                sortedIndicesi = sortedIndices[i]
                # delT(wij, wji, thetaij, M, Lambda0, L, frac, frac1, Mpfrac1, LambdaovM)
                wtilde = [w[i][j] + delT(w[i][j],
                                         w[j][i],
                                         Tstar[Tmap[(i,
                                                     j)]],
                                         M,
                                         Lambda0,
                                         L,
                                         frac,
                                         frac1,
                                         Mpfrac1,
                                         LambdaovM) / LCDCurrent[i] for j in sortedIndicesi]

                x = np.empty(shape=len(wtilde) + 1)
                # Solve the l1 projection problem.
                x[0] = utilde
                x[1:] = np.array(wtilde)
                projection = project(x)
                # Update the solution.
                u[i] = projection[0]
                # for j in range(len(w[i])):
                # w[i][sortedIndicesi[j]] = projection[j+1] ## +1 since u[i] is
                # first
                for counter, j in enumerate(sortedIndicesi):
                    w[i][j] = projection[counter + 1]
            # Calculate the current objective
            CDObj = CDObjConst  # 0
            for i in BfreeMinusZeroGroups:  # Bfree:
                CDObj += q(u[i], Bstar[Bmap[i]], M, Lambda0, L, frac)
                for j in w[i]:
                    if i < j:
                        # T(wij, wji, thetaij, M, Lambda0, L, frac, frac1, Mpfrac1, LambdaovM, Lambda1ovM)
                        CDObj += T(w[i][j], w[j][i], Tstar[Tmap[(i, j)]], M,
                                   Lambda0, L, frac, frac1, Mpfrac1, LambdaovM, Lambda1ovM)
            #Params["print"]("Inner obj: ", CDObj)
            if terminate(CDPrevObj, CDObj, TolCD):
                break
            CDPrevObj = CDObj

        # Get back the primal solution.
        for i in range(len(Bindices)):
            # if Bindices[i] is fixed to 1, then u[Bindices[i]] = 0 and the
            # update below will lead to currentB[i] = Bstar[i] (or +- M)
            if Bindices[i] not in zeroGroups:
                # assuming Bindices is sorted
                currentB[i] = dualtoprimalu(
                    u[Bindices[i]], Bstar[i], M, Lambda0, L, frac)
            else:
                currentB[i] = 0

        for i, j in Tmap:
            # if i or j is fixed, the corresponding w[i][j] will be zero, which
            # leads to the correct update.
            if i < j:
                if (i, j) in Tfree:
                    # dualtoprimalw(wij, wji, thetaij, M, Lambda0, L, frac, frac1, Mpfrac1)
                    if i in zeroGroups or j in zeroGroups:
                        currentT[Tmap[(i, j)]] = 0
                    else:
                        currentT[Tmap[(i, j)]] = dualtoprimalw(
                            w[i][j], w[j][i], Tstar[Tmap[(i, j)]], M, Lambda0, L, frac, frac1, Mpfrac1)
                else:  # careful, this is the case when no thresholding should be applied
                    coefficient = Tstar[Tmap[(i, j)]]
                    if np.abs(coefficient) <= M:
                        currentT[Tmap[(i, j)]] = coefficient
                    else:
                        currentT[Tmap[(i, j)]] = M * np.sign(coefficient)

        r = y - np.dot(XB, currentB) - np.dot(XT, currentT)

        maxterm = 0
        for i in range(len(currentB)):
            if Bindices[i] not in fixedB:
                maxtemp = np.abs(currentB[i])
                for j in w[Bindices[i]]:
                    maxtemp = max(maxtemp, np.abs(
                        currentT[Tmap[(Bindices[i], j)]]))
                maxterm += maxtemp
        l1norm = np.sum(np.abs(currentT[TfreeIndices]))
        # IMPORTANT: Avoid using lenFixed and lenFixedT here.....!!!!!! ####
        currentobjective = 0.5 * np.dot(r, r) + Lambda0 * (
            lenFixedB + lenFixedT) + (Lambda0 / M) * maxterm + (Lambda1 / M) * l1norm

        if currentobjective > oldObj:
            Params["print"]("Objective Increased!!!")

        if terminate(oldObj, currentobjective, Tol):
            break

        oldObj = currentobjective
        Params["print"]("Iteration :", it, ". Objective: ", currentobjective)

    end = time.time()
    Params["print"]("Time: ", end - start, " seconds.")

    # Check if any small values should be zero.
    # Start with more aggressive checks first.
    Trunc = False
    for epsilon in [0.01, 1e-3, 1e-4, 1e-5, 1e-6]:
        currentBtrunc = np.copy(currentB)
        currentTtrunc = np.copy(currentT)
        currentBSetToZero = np.nonzero(np.abs(currentB) < epsilon)[0]
        currentBtrunc[currentBSetToZero] = 0
        currentBSetToZeroPSet = set(currentBSetToZero)
        for (i, j) in Tmap:
            if Bmap[i] in currentBSetToZeroPSet or Bmap[j] in currentBSetToZeroPSet:
                currentTtrunc[Tmap[(i, j)]] = 0

        currentTtrunc[np.abs(currentT) < epsilon] = 0
        rtrunc = y - np.dot(XB, currentBtrunc) - np.dot(XT, currentTtrunc)
        maxterm = 0
        for i in range(len(currentBtrunc)):
            if Bindices[i] not in fixedB:
                maxtemp = np.abs(currentBtrunc[i])
                for j in w[Bindices[i]]:
                    maxtemp = max(maxtemp, np.abs(
                        currentTtrunc[Tmap[(Bindices[i], j)]]))
                maxterm += maxtemp
        l1norm = np.sum(np.abs(currentTtrunc[TfreeIndices]))
        objectivetrunc = 0.5 * np.dot(rtrunc, rtrunc) + Lambda0 * (
            lenFixedB + lenFixedT) + (Lambda0 / M) * maxterm + (Lambda1 / M) * l1norm

        Params["print"](
            "eps: ",
            epsilon,
            " objectivetrunc:  ",
            objectivetrunc,
            "  currentobjective: ",
            currentobjective)
        # 1.01 might be beneficial in some extreme cases where supp becomes
        # very large (but might also cause descent problems)
        if objectivetrunc <= currentobjective:
            '''
            currentB = currentBtrunc
            currentT = currentTtrunc
            r = rtrunc
            currentobjective = objectivetrunc
            '''
            Params["print"]("###CHANGE###", "eps: ", epsilon)
            Params["print"]("Final Objective :", objectivetrunc)
            Trunc = True
            break

    integral = True

    for i in Bfree:
        zi = np.abs(currentB[Bmap[i]]) / M
        if zi > 0 and zi < 0.999:
            integral = False

    for i in TfreeIndices:
        zi = np.abs(currentT[i]) / M
        if zi > 0 and zi < 0.999:
            integral = False

    Bnnz = {key: currentB[Bmap[key]]
            for key in Bmap if currentB[Bmap[key]] != 0}
    Tnnz = {(i, j): currentT[Tmap[(i, j)]]
            for i, j in Tmap if i < j and currentT[Tmap[(i, j)]] != 0}
    intercept = data.ybar - np.dot(XBMean, currentB) - np.dot(XTMean, currentT)
    sol = Solution(Bnnz, Tnnz, intercept)

    if Trunc:
        BnnzTrunc = {key: currentBtrunc[Bmap[key]]
                     for key in Bmap if currentBtrunc[Bmap[key]] != 0}
        TnnzTrunc = {(i, j): currentTtrunc[Tmap[(
            i, j)]] for i, j in Tmap if i < j and currentTtrunc[Tmap[(i, j)]] != 0}
        interceptTrunc = data.ybar - \
            np.dot(XBMean, currentBtrunc) - np.dot(XTMean, currentTtrunc)
        solTrunc = Solution(BnnzTrunc, TnnzTrunc, interceptTrunc)
    else:
        BnnzTrunc = Bnnz
        TnnzTrunc = Tnnz
        interceptTrunc = intercept
        solTrunc = sol

    return (sol, solTrunc, currentobjective, integral, r, u, w)



"""Module for fitting a regularization path."""

import time
from copy import deepcopy
from .InitialRun import InitialRun
from .Structs import *
from .FitSingleLambda import FitSingleLambda

def hier_fit(X, y, params):
    """
    Fits a regularization path for the interactions model under strong hierarchy.
    The grid of lambda_1's and lambda_2's is selected automatically.

    Inputs:
        X: Data matrix (numpy array) cotaining the main features. Interactions columns are
          generated on-the-fly and should not be included in X.
        y: Response vector (numpy array).
        params: A dictionary of parameters that control the fitting procedure.
          params supports the following options:
            "M": Float. Adds box constraints of the form |beta_i| <= M, |theta_{ij}| <= M.
              Default is 10^10.
            "alpha": Float. Used to define lambda_2: lambda_2 = alpha * lambda_1. Larger alphas
              lead to more sparsity in the interaction terms. To avoid performance issues,
              we recommend alpha >=1. Default is 2.
            "tol": Tolerance for PGD. Default 1e-6.
            "tolCD": Tolerance for Block Coordinate Ascent. Default is 1e-6.
            "nLambda": Number of solutions in the path. Default is 100.
            "lambdaMinRatio": A float < 1. The path is terminated at
              lambdaMinRatio*lambda_1_max, where lambda_1_max is the first lambda_1
              in the path (which sets all the main effects to zero). Smaller values
              will lead to denser solutions in the path. Default is 0.05.
            "maxSuppSize": The maximum number of non-zeros in the path after which to
              terminate.

    Outputs:
      A tuple (solutions, lambdas).
        solutions: A list of solutions. solutions[i] corresponds to the ith lambda
          in the path. solutions[i].B is a dictionary of the non-zero elements in
          B (the main coefficients vector) and solutions[i].T is a dictionary of the
          non-zero elements in T (the interaction effects vector). The intecept can
          be accessed as follows: solutions[i].intercept.
        lambdas: A list of the lambda_1's used in the path.
    """
    data = Matrices(X, y)
    Params = deepcopy(params)
    # Initialize the parameters not provided by the user.
    Params = InitializeParams(Params)
    # Store the print function in Params.
    Params["print"] = print if Params["debug"] else lambda *a, **k: None
    start = time.time()
    kmain = 5
    kint = 5
    # Get the initial active set and lambda_1 max.
    BActive, TActive, lambdaMax, maxNorm = InitialRun(
        data, kmain, kint, Params)
    # Generate the seq. of lambda_1's to be used in the path.
    Lambdas = np.geomspace(
        lambdaMax,
        Params["lambdaMinRatio"] *
        lambdaMax,
        num=Params["nLambda"])
    data.maxNorm = maxNorm  # used by CheckOptScreen.
    data.Augment(BActive, TActive)
    fixedVars = {}
    initialSol = Solution([], [])  # Change later when we start using this.
    u = {}
    w = {}
    useDual = False
    relaxationVars = RelaxationVars(
        BActive, TActive, initialSol, useDual, u, w)
    solutions = []
    for i in range(Params["nLambda"]):  # nLambda
        print("Path Iteration: ", i)
        Params["print"]("Lambda = ", Lambdas[i])
        Params["Lambda"] = Lambdas[i]
        # relaxationVars is updated by FitSingleLambda.
        sol, obj, integral = FitSingleLambda(
            Params, relaxationVars, fixedVars, data)
        solutions.append(deepcopy(sol))
        nnz = len(sol.B) + len(sol.T)
        if nnz >= Params["maxSuppSize"]:
            break
        Params["print"]("Number of nnz: ", nnz)
    end = time.time()
    Params["print"]("Total Elapsed Time: ", end - start)
    return solutions, Lambdas



"""Helper structs to store/maintain solutions and matrices."""

import math
import numpy as np
from scipy.special import comb


class Solution:
    '''
    B: dictionary. key = i
    T: dictionary. key = (i,j)
    '''

    def __init__(self, B, T, intercept=None, gap=None):
        # stores only non-zero indices/values.
        self.B = B.copy()
        self.T = T.copy()
        self.intercept = intercept
        self.gap = gap

    def ToArray(self, BIndices, TIndices):
        BArray = np.zeros(len(BIndices))
        for counter, index in enumerate(BIndices):
            if index in self.B:
                BArray[counter] = self.B[index]

        TArray = np.zeros(len(TIndices))
        for counter, index in enumerate(TIndices):
            if index in self.T:
                TArray[counter] = self.T[index]

        return (BArray, TArray)


class Matrices:
    def __init__(self, X, y, Bindices=[], Tindices=[], maxNorm=0):
        self.X = X
        self.y = y
        self.ybar = np.mean(self.y)
        self.ycentered = y - self.ybar
        # might need to change this in case materialization is used
        self.pMain = X.shape[1]
        self.P = comb(self.pMain, 2, exact=True)
        self.indexMapB = {Bindices[i]: i for i in range(
            len(Bindices))}  # maps var number to index
        self.indexMapT = {}  # maps var number to index
        self.XB = X[:, Bindices]
        self.XT = np.zeros(shape=(self.X.shape[0], len(Tindices)))
        l = 0
        for i, j in Tindices:
            self.XT[:, l] = X[:, i] * X[:, j]
            self.indexMapT[(i, j)] = l
            l += 1

        self.maxNorm = maxNorm

        self.bins = np.zeros(self.pMain, dtype=np.int64)
        for l in range(1, self.pMain):
            self.bins[l] = self.bins[l - 1] + self.pMain - l

        def interleave(p):
            for i in range(math.ceil(p / 2)):
                yield i
                yield p - i - 1

        self.balancedInd = list(interleave(self.pMain - 1))
        if len(self.balancedInd) != self.pMain - 1:
            self.balancedInd.pop()
        self.balancedInd = np.array(self.balancedInd)

    def Augment(self, Bindices, Tindices):
        BindicesNew = [i for i in Bindices if i not in self.indexMapB]
        TindicesNew = [key for key in Tindices if key not in self.indexMapT]

        pold = self.XB.shape[1]
        for i in range(len(BindicesNew)):
            self.indexMapB[BindicesNew[i]] = i + pold
        self.XB = np.hstack((self.XB, self.X[:, Bindices]))

        pold = self.XT.shape[1]
        XTtemp = np.zeros(shape=(self.X.shape[0], len(TindicesNew)))
        l = 0
        for i, j in TindicesNew:
            XTtemp[:, l] = self.X[:, i] * self.X[:, j]
            self.indexMapT[(i, j)] = l + pold
            l += 1
        self.XT = np.hstack((self.XT, XTtemp))

    def Retrieve(self, Bindices, Tindices):
        '''
        returns submatrices XB and XT corresponding to Bindices and Tindices
        '''
        Bidx = [self.indexMapB[i] for i in Bindices]
        Tidx = [self.indexMapT[i] for i in Tindices]
        return self.XB[:, Bidx], self.XT[:, Tidx]


class RelaxationVars:
    def __init__(
        self,
        BActive,
        TActive,
        initialSol,
        useDual,
        u,
        w,
        rold=np.zeros(1),
        maxrtX=None,
        Tviolating=set(),
            maxNorm=None):
        self.BActive = BActive  # the current active set -- type = list
        self.TActive = TActive  # the current active set -- type = list
        # initial sol to use for the convex solver -- type = Solution
        self.initialSol = initialSol
        self.u = u
        self.w = w
        self.useDual = useDual  # boolean to indicate whether to use a dual initialization
        self.rold = rold
        self.maxrtX = maxrtX
        self.Tviolating = Tviolating
        self.maxNorm = maxNorm


def FlatToTuple(x, data):
    """
    Converts a numpy array of flat indices to a python list of (i,j) pairs

    Derivation: (i+1)*(p-1) - i*(i-1)/2 + (j-i) - p = index => j = index - (i+1)*(p-1) + i*(i-1)/2 + i + p

    Test:
    for i in range(P):
        if FlatToTuple(np.array([i]))[0] != D[i + pMain]:
            Params["print"]("Problem!")
            Params["print"](i, FlatToTuple(np.array([i]))[0], D[i + pMain])
            break
    """
    iIndices = np.digitize(x, data.bins) - 1
    jIndices = x - (data.pMain - 1) * (iIndices + 1) + \
        iIndices * (iIndices - 1) / 2 + iIndices + data.pMain
    return list(zip(iIndices, jIndices.astype(int)))

def InitializeParams(params):
    default_params = {
        "M": 10**10,
        "alpha": 2,
        "tol": 1e-6,
        "tolCD": 1e-6,
        "CorrScreen": True,
        "nLambda": 100,
        "lambdaMinRatio": 0.05,
        "maxSuppSize": 500,
        "debug": False,
    }
    for key in default_params:
        if key not in params:
            params[key] = default_params[key]

    return params



"""Helper Utilities. Only contains the Predict function for now."""

import numpy as np


def hier_predict(sol, X):
    '''
    A predict function for hierScale.
    Inputs:
        sol: A single solution obtained from hier_fit.
        X: Matrix of (main effects) to predict the response.
    Output: A predictions vector.
    '''
    BIndices = sorted(sol.B.keys())
    TIndices = sorted(sol.T.keys())
    B, T = sol.ToArray(BIndices, TIndices)
    if len(TIndices) != 0:
        XTnew = np.hstack([(X[:, i] * X[:, j])[:, np.newaxis]
                           for (i, j) in TIndices])
        TPred = np.dot(XTnew, T)
    else:
        TPred = 0
    return np.dot(X[:, BIndices], B) + TPred + sol.intercept



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
     name='hierScale',
     version='1.0.6',
     author="Hussein Hazimeh and Rahul Mazumder",
     author_email="hazimeh@mit.edu",
     description="A scalable package for fitting sparse linear regression models with pairwise feature interactions, under strong hierarchy.",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/hazimehh/hierScale",
     packages=setuptools.find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
     python_requires='>=3.0',
     install_requires=["numpy", "scipy", "numba", "networkx"] # "gurobipy" is not available on pip.
 )
